"""Helper functions for dealing with FieldEdge classes and interservice comms.
"""
import inspect
import itertools
import json
import logging
import re

from fieldedge_utilities.logger import verbose_logging

__all__ = ['camel_case', 'snake_case', 'get_class_tag',
           'camel_to_snake', 'snake_to_camel',
           'get_class_properties', 'get_instance_properties_values',
           'json_compatible', 'hasattr_static',
           'property_is_read_only', 'property_is_async', 'tag_class_properties',
           'tag_class_property', 'untag_class_property', 'tag_merge',
           'equivalent_attributes', 'READ_ONLY', 'READ_WRITE']

READ_ONLY = 'info'
READ_WRITE = 'config'

_log = logging.getLogger(__name__)


def camel_to_snake(camel_str: str, skip_caps: bool = False) -> str:
    """Converts a camelCase string to snake_case.
    
    **DEPRECATED** use `snake_case` instead.

    Args:
        camel_str: The string to convert.
        skip_caps: A flag if `True` will return CAPITAL_CASE unchanged.
        
    Returns:
        The input string in snake_case format.
        
    Raises:
        `ValueError` if camel_str is not a valid string.
        
    """
    return snake_case(camel_str, skip_caps, skip_pascal=True)


def snake_to_camel(snake_str: str, skip_caps: bool = False) -> str:
    """Converts a snake_case string to camelCase.
    
    **DEPRECATED** use `camel_case` instead.
    
    Args:
        snake_str: The string to convert.
        skip_caps: If `True` will return CAPITAL_CASE unchanged
    
    Returns:
        The input string in camelCase structure.
        
    """
    return camel_case(snake_str, skip_caps, skip_pascal=True)


def snake_case(original: str,
               skip_caps: bool = False,
               skip_pascal: bool = False) -> str:
    """Converts a string to snake_case.
    
    Args:
        original: The string to convert.
        skip_caps: A flag if `True` will return CAPITAL_CASE unchanged.
        skip_pascal: A flag if `True` will return PascalCase unchanged.
        
    Returns:
        The original string converted to snake_case format.
        
    Raises:
        `ValueError` if original is not a valid string.
        
    """
    if not isinstance(original, str) or not original:
        raise ValueError('Invalid string input')
    if original.isupper() and skip_caps:
        return original
    snake = re.compile(r'(?<!^)(?=[A-Z])').sub('_', original).lower()
    if '__' in snake:
        words = snake.split('__')
        snake = '_'.join(f'{word.replace("_", "")}' for word in words)
    words = snake.split('_')
    if original[0].isupper() and skip_pascal:
        if all(word.title() in original for word in words):
            return original
    return snake


def camel_case(original: str,
               skip_caps: bool = False,
               skip_pascal: bool = False) -> str:
    """Converts a string to camelCase.
    
    Args:
        original: The string to convert.
        skip_caps: If `True` will return CAPITAL_CASE unchanged
        skip_pascal: If `True` will return PascalCase unchanged
    
    Returns:
        The input string in camelCase structure.
        
    """
    if not isinstance(original, str) or not original:
        raise ValueError('Invalid string input')
    if original.isupper() and skip_caps:
        return original
    words = original.split('_')
    if len(words) == 1:
        regex = '.+?(?:(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])|$)'
        matches = re.finditer(regex, original)
        words = [m.group(0) for m in matches]
    if skip_pascal and all(word.title() == word for word in words):
        return original
    return words[0].lower() + ''.join(w.title() for w in words[1:])


def pascal_case(original: str, skip_caps: bool = False) -> str:
    """Returns the string converted to PascalCase.
    
    Args:
        original: The original string.
        skip_caps: A flag that returns the original if CAPITAL_CASE.

    """
    camel = camel_case(original, skip_caps)
    return camel[0].upper() + camel[1:]


def get_class_tag(cls: type) -> str:
    """Returns a lowercase name to use as the tag for a class."""
    if isinstance(cls, type):
        return cls.__name__.lower()
    return cls.__class__.__name__.lower()


def get_class_properties(cls: type, ignore: 'list[str]' = None) -> 'list[str]':
    """Returns non-hidden, non-callable properties/values of a Class instance.
    
    Also ignores CAPITAL_CASE attributes which are assumed to be constants.
    
    Args:
        cls: The Class whose properties will be derived
        ignore: A list of names to ignore (optional)
    
    Returns:
        A list of exposed property names.
        
    Raises:
        ValueError if `cls` does not have a `dir()` method or is not a `type`.
        
    """
    if not dir(cls):
        raise ValueError('Invalid cls_instance - must have dir() method')
    if isinstance(cls, type) and '__slots__' not in dir(cls):
        _log.warning('No __slots__: attributes in __init__ will be missed')
    if not isinstance(ignore, list):
        ignore = []
    attrs = [attr for attr in dir(cls)
             if not attr.startswith(('_',)) and
             attr not in ignore and
             not callable(inspect.getattr_static(cls, attr)) and
             not attr.isupper()]
    return attrs


def get_instance_properties_values(instance: object) -> dict:
    """Returns the instance properties and values."""
    props_list = get_class_properties(instance)
    props_values = {}
    for prop in props_list:
        props_values[prop] = getattr(instance, prop)
    return props_values


def json_compatible(obj: object,
                    camel_keys: bool = True,
                    skip_caps: bool = True) -> dict:
    """Returns a dictionary compatible with `json.dumps` function.

    Nested objects are converted to dictionaries.
    
    `LOG_VERBOSE` optional key: `tags`
    
    Args:
        obj: The source object.
        camel_keys: Flag indicating whether to convert all nested dictionary
            keys to `camelCase`.
        skip_caps: Preserves `CAPITAL_CASE` keys if True
        
    Returns:
        A dictionary with nested arrays, dictionaries and other compatible with
            `json.dumps`.

    """
    res = obj
    if camel_keys:
        if isinstance(obj, dict):
            res = {}
            for key, val in obj.items():
                if ((isinstance(key, str) and key.isupper() and skip_caps) or
                    not isinstance(key, str)):
                    # no change
                    camel_key = key
                else:
                    camel_key = camel_case(str(key))
                if camel_key != key and verbose_logging('tags'):
                    _log.debug('Changed %s to %s', key, camel_key)
                res[camel_key] = json_compatible(val, camel_keys, skip_caps)
        elif isinstance(obj, list):
            res = [json_compatible(i) for i in obj]
    try:
        json.dumps(res)
        return res
    except TypeError:
        try:
            if callable(res):
                res = f'<function:{res.__name__}>'
            elif isinstance(res, list):
                res = [json_compatible(v, camel_keys, skip_caps)
                       for v in res]
            elif hasattr(res, '__dict__'):
                res = json_compatible(get_instance_properties_values(res),
                                      camel_keys,
                                      skip_caps)
            elif isinstance(res, dict):
                res = {k:json_compatible(v, camel_keys, skip_caps)
                       for k, v in res.items()}
            elif hasattr(res, '__slots__'):
                res = {s: json_compatible(getattr(res, s, None))
                       for s in res.__slots__}
            else:
                res = '<non-serializable>'
            return res
        except Exception as exc:
            _log.error(exc)
            raise exc


def hasattr_static(obj: object, attr: str) -> bool:
    """Determines if an object has an attribute without calling the attribute.
    
    Args:
        obj: The object to inspect.
        attr: The name of the attribute to query.
    
    Returns:
        `True` if the object has the attribute.
        
    """
    try:
        inspect.getattr_static(obj, attr)
        return True
    except AttributeError:
        return False


def property_is_read_only(instance: object, property_name: str) -> bool:
    """Returns True if the instance attribute has no fset method."""
    if not hasattr_static(instance, property_name):
        raise ValueError(f'Object has no property {property_name}')
    prop = inspect.getattr_static(instance, property_name)
    try:
        return prop.fset is None
    except AttributeError:
        return False


def property_is_async(instance: object, property_name: str) -> bool:
    """Returns True if an object is awaitable."""
    if not hasattr_static(instance, property_name):
        raise ValueError(f'Object has no property {property_name}')
    return inspect.isawaitable(getattr(instance, property_name))


def tag_class_properties(cls: type,
                         tag: str = None,
                         auto_tag: bool = True,
                         use_json: bool = True,
                         categorize: bool = False,
                         ignore: 'list[str]' = None,
                         ) -> 'list|dict':
    """Retrieves the class public properties tagged with a routing prefix.
    
    If a `tag` is not provided and `auto_tag` is `True` then the lowercase name
    of the instance's class will be used e.g. MyClass.property becomes
    myclassProperty.
    
    Using the defaults will return a simple list of tagged property names
    with the form `['tagProp1Name', 'tagProp2Name']`
    
    If `tag` is `None` and `auto_tag` is `False` then no tag will be applied
    and the native property names will be returned as JSON if `json` is `True`.
    
    If `categorize` is `True` a dictionary is returned of the form
    `{ 'info': ['tagProp1Name'], 'config': ['tagProp2Name']}` where
    the category is not present if no properties meet the respective criteria.
    
    If `json` is `False` the above applies but property names will use
    their original case e.g. `tag_prop1_name`
    
    `LOG_VERBOSE` optional key: `tags`.
    
    Args:
        cls: A class to tag.
        tag: The name of the routing prefix. If `None`, the calling function's
            module `__name__` will be used.
        auto_tag: If `True` will use the class name in lowercase.
        json: A flag indicating whether to use camelCase keys.
        categorize: A flag indicating whether to group as `info` and `config`.
        ignore: A list of property names to ignore.
    
    Retuns:
        A dictionary or list of strings (see docstring).
        
    """
    # TODO: class checking seems not to work for certain subclasses
    if isinstance(cls, type) and verbose_logging('tags'):
        _log.debug('Processing for class type')
    # elif issubclass(cls, ABC):
    #     _log.debug('Processing for microservice')
    if auto_tag and not tag:
        tag = get_class_tag(cls)
    class_props = get_class_properties(cls, ignore)
    if not categorize:
        return [tag_class_property(prop, tag, use_json) for prop in class_props]
    result = {}
    for prop in class_props:
        if property_is_read_only(cls, prop):
            if READ_ONLY not in result:
                result[READ_ONLY] = []
            result[READ_ONLY].append(tag_class_property(prop, tag, use_json))
        else:
            if READ_WRITE not in result:
                result[READ_WRITE] = []
            result[READ_WRITE].append(tag_class_property(prop, tag, use_json))
    return result


def tag_class_property(prop: str,
                       tag_or_cls: 'str|type' = None,
                       use_json: bool = True) -> str:
    """Converts a property for ISC adding an optional tag."""
    if tag_or_cls is None:
        tagged = prop
    else:
        if isinstance(tag_or_cls, type):
            tag = get_class_tag(tag_or_cls)
        elif isinstance(tag_or_cls, str):
            tag = tag_or_cls
        else:
            raise ValueError('tag_or_cls must be a string or class type')
        tagged = f'{tag.lower()}_{prop}'
    if use_json:
        return camel_case(f'{tagged}')
    return f'{tag}_{prop}'


def untag_class_property(property_name: str,
                         is_tagged: bool = True,
                         include_tag: bool = False,
                         ) -> 'str|tuple[str, str]':
    """Reverts a JSON-format tagged property to its PEP representation.
    
    Expects a JSON-format tagged value e.g. `modemUniqueId` would return
    `(unique_id, modem)` where it assumes the first word is the tag.

    Args:
        property_name: The property name, assumes using camelCase.
        include_tag: If True, a tuple is returned with the tag as the second
            element.
    
    Returns:
        A string with the original property name, or a tuple with the original
            property value in snake_case, and the tag

    """
    prop = snake_case(property_name)
    tag = None
    if is_tagged:
        if '_' not in prop:
            raise ValueError(f'Invalid tagged {property_name}')
        tag, prop = prop.split('_', 1)
    if not include_tag:
        return prop
    return (prop, tag)


def tag_merge(*args) -> 'list|dict':
    """Merge multiple tagged property lists/dictionaries.
    
    Args:
        *args: A set of dictionaries or lists, must all be the same structure.
    
    Returns:
        Merged structure of whatever was passed in.

    """
    container_type = args[0].__class__.__name__
    if container_type not in ('list', 'dict'):
        raise ValueError('tag merge must be of list or dict type')
    if not all(arg.__class__.__name__ == container_type for arg in args):
        raise ValueError('args must all be of same type')
    if container_type == 'list':
        return list(itertools.chain(*args))
    merged = {}
    categories = [READ_ONLY, READ_WRITE]
    dict_0: dict = args[0]
    if any(k in categories for k in dict_0):
        for arg in args:
            assert isinstance(arg, dict)
            if not any(k in categories for k in arg):
                raise ValueError('Not all dictionaries are categorized')
            merged = _nested_tag_merge(arg, merged)
    else:
        for arg in args:
            assert isinstance(arg, dict)
            for key, val in arg.items():
                merged[key] = val
    return merged


def _nested_tag_merge(add: dict, merged: dict) -> dict:
    for key, val in add.items():
        if key not in merged:
            merged[key] = val
        else:
            if isinstance(merged[key], list):
                merged[key] = merged[key] + val
            else:
                assert isinstance(merged[key], dict)
                assert isinstance(val, dict)
                for nested_key, nested_val in val.items():
                    merged[key][nested_key] = nested_val
    return merged


def equivalent_attributes(ref: object,
                          other: object,
                          exclude: 'list[str]' = None,
                          dbg: str = '',
                          ) -> bool:
    """Confirms attribute equivalence between objects of the same type.
    
    Args:
        ref: The reference object being compared to.
        other: The object comparing against the reference.
        exclude: Optional list of attribute names to exclude from comparison.
    
    Returns:
        True if all (non-excluded) attribute name/values match.

    """
    if not isinstance(other, type(ref)):
        return False
    if not hasattr(ref, '__dict__') or not hasattr(other, '__dict__'):
        return ref == other
    if not isinstance(exclude, list):
        exclude = []
    if dbg:
        dbg += '.'
    for attr in dir(ref):
        if attr.startswith('__') or attr in exclude:
            continue
        if not hasattr(other, attr):
            _log.debug('Other missing %s%s', dbg, attr)
            return False
        ref_val = getattr(ref, attr)
        if callable(ref_val):
            continue
        other_val = getattr(other, attr)
        if any(hasattr(ref_val, a) for a in ['__dict__', '__slots__']):
            if not equivalent_attributes(ref_val, other_val, dbg=attr):
                return False
        elif ref_val != other_val:
            _log.debug('%s%s mismatch', dbg, attr)
            return False
    return True



"""A proxy class for interfacing with other Microservices via MQTT.
"""
import logging
import os
from abc import ABC, abstractmethod
from enum import IntEnum
from threading import Event
from typing import Callable, Any

from fieldedge_utilities.timer import RepeatingTimer
from fieldedge_utilities.logger import verbose_logging

from .interservice import IscTaskQueue, IscTask, IscException
from .propertycache import PropertyCache

__all__ = ['MicroserviceProxy', 'InitializationState']

_log = logging.getLogger(__name__)

PROXY_PROPERTY_TIMEOUT = int(os.getenv('PROXY_PROPERTY_TIMEOUT', '35'))


class InitializationState(IntEnum):
    """Initialization state of the MicroserviceProxy."""
    NONE = 0
    PENDING = 1
    COMPLETE = 2


class MicroserviceProxy(ABC):
    """A proxy model for another FieldEdge microservice accessed via MQTT.
    
    Queries a microservice based on its tag to populate proxy_properties.
    Has a blocking (1-deep) `IscTaskQueue` for each remote query to complete
    before the next task can be queued.
    
    """
    def __init__(self, **kwargs):
        """Initialize the proxy.
        
        Keyword Args:
            tag (str): The name of the microservice used in the MQTT topic.
                If not provided will use the lowercase class name.
            publish (Callable[[str, dict]]): Parent MQTT publish function
            subscribe (Callable[[str]]): Parent MQTT subscribe function
            unsubscribe (Callable[[str]]): Parent MQTT unsubscribe function
            init_callback (Callable[[bool, str]]): Optional callback when
                initialize() completes.
            init_timeout (int): Time in seconds allowed for initialization.
            cache_lifetime (int): The proxy property cache time.
            isc_poll_interval (int): The time between checks for task expiry.
        
        """
        self._tag: str = (kwargs.get('tag', None) or
                          self.__class__.__name__.lower())
        if not self._tag:
            raise ValueError('Invalid tag provided')
        callbacks = ['publish', 'subscribe', 'unsubscribe', 'init_callback']
        int_config = ['init_timeout', 'cache_lifetime', 'isc_poll_interval']
        for key, val in kwargs.items():
            if key in callbacks:
                if not callable(val):
                    raise ValueError(f'{key} must be callable')
            elif key in int_config:
                if not isinstance(val, int) or val <= 0:
                    raise ValueError(f'{key} must be integer > 0')
        self._publish: Callable[[str, dict], None] = kwargs.get('publish', None)
        self._subscribe: Callable[['str|list[str]'], bool] = (
            kwargs.get('subscribe', None))
        self._unsubscribe: Callable[['str|list[str]'], bool] = (
            kwargs.get('unsubscribe', None))
        self._init_callback: Callable[[bool, str], None] = (
            kwargs.get('init_callback', None))
        self._init_timeout: int = kwargs.get('init_timeout', 10)
        self._cache_lifetime: 'int|None' = kwargs.get('cache_lifetime', None)
        self._isc_poll_interval: int = kwargs.get('isc_poll_interval', 1)
        self.isc_queue = IscTaskQueue(blocking=True)
        self._isc_timer = RepeatingTimer(seconds=self._isc_poll_interval,
                                         target=self.isc_queue.remove_expired,
                                         name='IscTaskExpiryTimer',
                                         auto_start=True)
        self._proxy_properties: dict = None
        self._property_cache: PropertyCache = PropertyCache()
        self._proxy_event: Event = Event()
        self._init: InitializationState = InitializationState.NONE

    @property
    def tag(self) -> str:
        """The name of the microservice used in MQTT topic."""
        return self._tag

    @property
    def is_initialized(self) -> bool:
        """Returns True if the proxy has been initialized with properties."""
        return self._init == InitializationState.COMPLETE

    @property
    def initialization_state(self) -> InitializationState:
        """The current initialization state."""
        return self._init

    @property
    def _base_topic(self) -> str:
        if not self.tag:
            raise ValueError('tag is not defined')
        return f'fieldedge/{self.tag}'

    @property
    def properties(self) -> 'dict|None':
        """The microservice properties.
        
        If cached returns immediately, otherwise blocks waiting for an update
        via the MQTT thread. Some properties e.g. GNSS information may take
        longer than 30 seconds to resolve.
        
        Raises:
            `OSError` if the proxy has not been initialized, or if the request
            times out after `PROXY_PROPERTY_TIMEOUT` seconds (default 35).
        
        """
        if self._init < InitializationState.PENDING:
            raise OSError('Proxy not initialized')
        cached = self._property_cache.get_cached('all')
        if cached:
            return self._proxy_properties
        pending = self.isc_queue.peek(task_meta=('properties', 'all'))
        if pending:
            _log.debug('Prior query pending')
        else:
            self._proxy_properties = None
            task_meta = { 'properties': 'all' }
            if self._proxy_event.is_set():
                self._proxy_event.clear()
            self.query_properties(['all'], task_meta)
        self._proxy_event.wait(PROXY_PROPERTY_TIMEOUT)
        if not self._proxy_properties:
            raise OSError('proxy_properties unsuccessful')
        return self._proxy_properties

    def property_get(self, property_name: str) -> Any:
        """Gets the proxy property value."""
        cached = self._property_cache.get_cached(property_name)
        if cached:
            return cached
        return self.properties.get(property_name)

    def property_set(self, property_name: str, value: Any, **kwargs):
        """Sets the proxy property value."""
        task_meta = { 'set': property_name }
        self.query_properties({ property_name: value }, task_meta, kwargs)

    def task_add(self, task: IscTask) -> None:
        """Adds a task to the task queue."""
        self.isc_queue.task_blocking.wait()
        try:
            self.isc_queue.append(task)
        except IscException as err:
            self.isc_queue.task_blocking.set()
            raise err

    def task_handle(self, response: dict, unblock: bool = False) -> bool:
        """Returns True if the task was handled, after triggering any callback.
        
        Args:
            response (dict): The response message from the microservice.
        
        """
        task_id = response.get('uid', None)
        if not task_id or not self.isc_queue.is_queued(task_id):
            _log.debug('No task ID %s queued - ignoring', task_id)
            return False
        task = self.isc_queue.get(task_id, unblock=unblock)
        if not isinstance(task.task_meta, dict):
            if task.task_meta is not None:
                _log.warning('Overwriting task_meta: %s', task.task_meta)
            task.task_meta = {}
        task.task_meta['task_id'] = task_id
        task.task_meta['task_type'] = task.task_type
        if callable(task.callback):
            task.callback(response, task.task_meta)
        elif self.isc_queue.task_blocking:
            _log.warning('Task queue still blocking with no callback')
        return True

    def task_complete(self, task_meta: dict = None):
        """Call to complete a task and remove from the blocking queue."""
        task_id = None
        if isinstance(task_meta, dict):
            task_id = task_meta.get('task_id', None)
            task_type = task_meta.get('task_type', 'task')
        _log.debug('Completing %s (%s)', task_type, task_id)
        self.isc_queue.task_blocking.set()

    def initialize(self, **kwargs) -> None:
        """Requests properties of the microservice to create the proxy."""
        topics = [f'{self._base_topic}/event/#', f'{self._base_topic}/info/#']
        for topic in topics:
            if callable(self._subscribe):
                subscribed = self._subscribe(topic)
                if not subscribed:
                    raise ValueError(f'Unable to subscribe to {topic}')
        task_meta = {
            'initialize': self.tag,
            'timeout': self._init_timeout,
            'timeout_callback': self._init_fail,
        }
        self._init = InitializationState.PENDING
        self.query_properties(['all'], task_meta, kwargs)

    def deinitialize(self) -> None:
        """De-initialize the proxy and clear the property cache and task queue.
        """
        self._init = InitializationState.NONE
        self._property_cache.clear()
        self.isc_queue.clear()

    def _init_fail(self, task_meta: dict = None):
        """Calls back with a failure on initialization failure/timeout."""
        self._init = InitializationState.NONE
        if callable(self._init_callback):
            tag = None
            if isinstance(task_meta, dict):
                tag = task_meta.get('initialize', None)
            self._init_callback(success=False, tag=tag)

    def query_properties(self,
                         properties: 'dict|list',
                         task_meta: dict = None,
                         query_meta: dict = None):
        """Gets or sets the microservice properties via MQTT.
        
        Args:
            properties: A list for `get` or a dictionary for `set`.
            task_meta: Optional dictionary elements for cascaded functions.
            query_meta: Optional metadata to add to the MQTT message query.
            
        """
        if not callable(self._publish):
            raise ValueError('publish callback not defined')
        if properties is not None and not isinstance(properties, (list, dict)):
            raise ValueError('Invalid properties structure')
        if isinstance(properties, dict):
            if not properties:
                raise ValueError('Properties dictionary must include key/values')
            method = 'set'
        else:
            method = 'get'
        _log.debug('%sting %s properties %s', method, self.tag, properties)
        if isinstance(task_meta, dict):
            lifetime = task_meta.get('timeout', 10)
        else:
            lifetime = 10
        prop_task = IscTask(task_type=f'property_{method}',
                            task_meta=task_meta,
                            callback=self.update_proxy_properties,
                            lifetime=lifetime)
        self.task_add(prop_task)
        topic = f'{self._base_topic}/request/properties/{method}'
        message = {
            'uid': prop_task.uid,
            'properties': properties,
        }
        if isinstance(query_meta, dict):
            for key, val in query_meta.items():
                message[key] = val
        self._publish(topic, message)

    def update_proxy_properties(self, message: dict, task_meta: dict = None):
        """Updates the proxy property dictionary with queried values.
        
        If querying all properties, pushes values into warm storage under
        self._proxy_properties. Otherwise hot storage in self._property_cache.
        """
        properties = message.get('properties', None)
        if not isinstance(properties, dict):
            _log.error('Unable to process properties: %s', properties)
            return
        cache_lifetime = self._cache_lifetime
        cache_all = False
        new_init = False
        if isinstance(task_meta, dict):
            if 'initialize' in task_meta:
                self._init = InitializationState.COMPLETE
                new_init = True
                cache_all = True
                _log.info('%s proxy initialized', self.tag)
            if 'cache_lifetime' in task_meta:
                cache_lifetime = task_meta.get('cache_liftime')
            if task_meta.get('properties', None) == 'all':
                cache_all = True
        if self._proxy_properties is None:
            self._proxy_properties = {}
        for prop, val in properties.items():
            if (prop not in self._proxy_properties or
                self._proxy_properties[prop] != val):
                _log.debug('Updating %s = %s', prop, val)
                self._proxy_properties[prop] = val
                self._property_cache.cache(val, prop, cache_lifetime)
        if cache_all:
            self._property_cache.cache(cache_all, 'all', cache_lifetime)
            if not self._proxy_event.is_set():
                self._proxy_event.set()
        self.task_complete(task_meta)
        if new_init and callable(self._init_callback):
            self._init_callback(success=True,
                                tag=task_meta.get('initialize', None))

    def publish(self, topic: str, message: dict, qos: int = 0):
        """Publishes to MQTT via the parent."""
        if not callable(self._publish):
            raise ValueError('publish callback not defined')
        self._publish(topic, message, qos=qos)

    def subscribe(self, topic: str):
        """Subscribes to a MQTT topic via the parent."""
        if not callable(self._subscribe):
            raise ValueError('subscribe callback not defined')
        self._subscribe(topic)

    def unsubscribe(self, topic: str):
        """Subscribes to a MQTT topic via the parent."""
        if not callable(self._unsubscribe):
            raise ValueError('unsubscribe callback not defined')
        self._unsubscribe(topic)

    @abstractmethod
    def on_isc_message(self, topic: str, message: dict) -> bool:
        """Processes MQTT messages for the proxy.
        
        Required method. Called by the parent's MQTT message handler.
        Should call self.task_handle if info/properties/values is received.
        Should return False by default, True if the message was handled.
        
        Args:
            topic (str): The message topic.
            message (dict): The message content.
        
        Returns:
            `True` if the message was processed or `False` otherwise.
            
        """
        if _vlog(self.tag):
            _log.debug('Proxy received %s: %s', topic, message)
        if not topic.startswith(f'fieldedge/{self.tag}/'):
            return False
        if topic.endswith('info/properties/values'):
            return self.task_handle(message)
        return False


def _vlog(tag: str) -> bool:
    return verbose_logging(f'{tag}-msproxy')



"""Classes for managing satellite modem interaction and metrics.

`ConnectionManager` provides housekeeping for attempting to connect to a
modem that may have transient connectivity to the microcontroller.
It allows for backing off from retrying connection, counting timeouts and CRC
errors from AT command responses and threshold maximums for each.

`QosMetricsManager` provides housekeeping for sample and metrics publishing
intervals that are interdependent.

"""
import logging
from dataclasses import dataclass

__all__ = ['ConnectionManager', 'QosMetricsManager']

_log = logging.getLogger(__name__)


@dataclass
class ConnectionManager:
    """Counters for connection management.
    
    The invoking application is expected to increment the counters accordingly.
    
    """
    retry_interval: int = 30
    backoff_interval: int = 30
    init_attempts: int = 0
    at_timeouts: int = 0
    max_at_timeouts: int = 3
    at_crc_errors: int = 0
    max_crc_errors: int = 5

    def __post_init__(self):
        if self.backoff_interval != self.retry_interval:
            self.backoff_interval = self.retry_interval

    def timeouts_exceeded(self) -> bool:
        """Returns True if at_timeouts >= max_at_timeouts."""
        return self.at_timeouts >= self.max_at_timeouts

    def timeout_increment_check(self) -> bool:
        """Increments at_timeouts and returns True if threshold exceeded."""
        self.at_timeouts += 1
        return self.timeouts_exceeded()

    def crc_exceeded(self) -> bool:
        """Returns True if at_crc_errors >= max_crc_errors."""
        return self.at_crc_errors >= self.max_crc_errors

    def crc_increment_check(self) -> bool:
        """Increments at_crc_erros and returns True if threshold exceeded."""
        self.at_crc_errors += 1
        return self.crc_exceeded()

    def reset(self) -> None:
        """Resets all counters to zero."""
        self.init_attempts = 0
        self.at_timeouts = 0
        self.at_crc_errors = 0
        self.backoff_interval = self.retry_interval

    def backoff(self) -> None:
        """Increases the backoff_interval for initialization retries.
        
        Every 3 attempts, the backoff_interval will double up to a maximum
        of one day (86400 seconds).
        
        """
        if self.init_attempts > 3 and self.init_attempts % 3 == 0:
            if self.backoff_interval * 2 <= 86400:
                self.backoff_interval = self.backoff_interval * 2


class QosMetricsManager:
    """Manages settings and counters for QoS metrics.
    
    The invoking application is expected to increment the counters accordingly.
    
    """
    def __init__(self,
                 sample_interval: int = 12,
                 metrics_interval: int = 300) -> None:
        """Initializes the metrics manager."""
        self._metrics_interval: int = 300
        self.metrics_interval = metrics_interval
        self._sample_interval: int = 12
        self.sample_interval = sample_interval
        self.sample_count: int = 0
        self.metrics_count: int = 0
        self.last_sample_ts: int = 0
        self.last_metrics_ts: int = 0

    @property
    def metrics_interval(self) -> int:
        """The metrics publish interval in seconds. (0..86400)"""
        return self._metrics_interval

    @metrics_interval.setter
    def metrics_interval(self, value: int):
        if not isinstance(value, int) or value not in range (0, 86401):
            raise ValueError('value must be integer 0..86400')
        self._metrics_interval = value

    def reset_counters(self) -> None:
        """Resets the metrics counters."""
        self.metrics_count = 0
        self.sample_count = 0

    @property
    def sample_interval(self) -> int:
        """The metrics sample interval in seconds. (0..86400)"""
        return self._sample_interval

    @sample_interval.setter
    def sample_interval(self, value: int):
        if (not isinstance(value, int) or
            value < 0 or
            value > self.metrics_interval):
            # Invalid
            raise ValueError('value must be integer 0..metrics_interval')
        if value > 0 and self.metrics_interval % value != 0:
            value = int(self.metrics_interval / value)
            _log.warning('Adjusted sample_interval to %d', value)
        self._sample_interval = value

    def sample_increment_check(self,
                               reset_sample_count: bool = False,
                               increment_metrics_count: bool = False) -> bool:
        """Increments the sample counter and returns True if metrics are due.
        
        Sample and metrics counters can be managed by setting flags in the
        method arguments.
        
        Args:
            reset_sample_count: If `True` and metrics are due, sample_count is
                reset to zero.
            increment_metrics_count: If `True` and metrics are due,
                metrics_count is incremented.
        
        """
        self.sample_count += 1
        count_threshold = int(self.metrics_interval / self.sample_interval)
        metrics_due = self.sample_count % count_threshold == 0
        if metrics_due:
            if reset_sample_count:
                self.sample_count = 0
            if increment_metrics_count:
                self.metrics_count += 1
        return metrics_due



"""
.. include:: ../../docs/overviews/microservice.md
"""
from .feature import Feature
from .interservice import IscException, IscTask, IscTaskQueue
from .microservice import Microservice
from .propertycache import PropertyCache
from .msproxy import MicroserviceProxy
from .subscriptionproxy import SubscriptionProxy

__all__ = ['Feature', 'IscException', 'IscTask', 'IscTaskQueue', 'Microservice',
           'PropertyCache', 'MicroserviceProxy', 'SubscriptionProxy']
